@@ -54,6 +54,7 @@
 #include <libtracker-extract/tracker-extract.h>
 
 #include "tracker-cue-sheet.h"
+#include "tracker-main.h"
 
 typedef enum {
 	EXTRACT_MIME_AUDIO,
@@ -104,6 +105,8 @@ typedef struct {
 #endif
 } MetadataExtractor;
 
+static TrackerSparqlConnection *local_conn = NULL;
+
 static void common_extract_stream_metadata (MetadataExtractor    *extractor,
                                             const gchar          *uri,
                                             TrackerResource      *resource);
@@ -690,7 +693,7 @@ extractor_apply_audio_metadata (MetadataExtractor     *extractor,
 			tracker_resource_set_gvalue (hash_resource, "nfo:hashValue", &acoustid_fingerprint);
 			tracker_resource_set_string (hash_resource, "nfo:hashAlgorithm", "chromaprint");
 
-			tracker_resource_add_take_relation (file_resource, "nfo:hasHash", hash_resource);
+			tracker_resource_set_take_relation (file_resource, "nfo:hasHash", hash_resource);
 			g_value_unset (&acoustid_fingerprint);
 		}
 	#endif
@@ -849,10 +852,15 @@ extract_metadata (MetadataExtractor      *extractor,
                   const gchar            *file_url)
 {
 	TrackerResource *resource;
+	gchar *resource_uri;
+	GFile *file;
 
 	g_return_val_if_fail (extractor != NULL, NULL);
 
-	resource = tracker_resource_new (NULL);
+	file = g_file_new_for_uri (file_url);
+	resource_uri = tracker_file_get_content_identifier (file, NULL, "1");
+	resource = tracker_resource_new (resource_uri);
+	g_free (resource_uri);
 
 	if (extractor->toc) {
 		gst_tag_list_insert (extractor->tagcache,
@@ -974,10 +982,18 @@ extract_metadata (MetadataExtractor      *extractor,
 					/* Reuse the "root" InformationElement resource for the first track,
 					 * so there's no spare ones.
 					 */
-					if (node == extractor->toc->entry_list)
+					if (node == extractor->toc->entry_list) {
 						track = resource;
-					else
-						track = tracker_resource_new (NULL);
+					} else {
+						gchar *suffix, *resource_uri;
+
+						suffix = g_strdup_printf ("%d", g_list_position (extractor->toc->entry_list,
+						                                                 node) + 1);
+						resource_uri = tracker_file_get_content_identifier (file, NULL, suffix);
+						track = tracker_resource_new (resource_uri);
+						g_free (resource_uri);
+						g_free (suffix);
+					}
 
 					extract_track (track, extractor, node->data, file_url, album_disc);
 					tracker_resource_set_relation (track, "nie:isStoredAs", file_resource);
@@ -1034,6 +1050,8 @@ extract_metadata (MetadataExtractor      *extractor,
 
 	common_extract_stream_metadata (extractor, file_url, resource);
 
+	g_object_unref (file);
+
 	return resource;
 }
 
@@ -1229,7 +1247,18 @@ discoverer_init_and_run (MetadataExtractor *extractor,
 	extractor->duration = gst_discoverer_info_get_duration (info) / GST_SECOND;
 
 	/* Retrieve global tags */
+#if defined(HAVE_GSTREAMER_1_20)
+	GstDiscovererStreamInfo *sinfo = gst_discoverer_info_get_stream_info (info);
+
+	if (GST_IS_DISCOVERER_CONTAINER_INFO (sinfo))
+		discoverer_tags = gst_discoverer_container_info_get_tags (GST_DISCOVERER_CONTAINER_INFO (sinfo));
+	else if (GST_IS_DISCOVERER_STREAM_INFO (sinfo))
+		discoverer_tags = gst_discoverer_stream_info_get_tags (sinfo);
+	else
+		discoverer_tags = NULL;
+#else
 	discoverer_tags = gst_discoverer_info_get_tags (info);
+#endif
 
 	if (discoverer_tags) {
 		gst_tag_list_insert (extractor->tagcache,
@@ -1241,7 +1270,6 @@ discoverer_init_and_run (MetadataExtractor *extractor,
 	extractor->streams = gst_discoverer_info_get_stream_list (info);
 	for (l = extractor->streams; l; l = g_list_next (l)) {
 		GstDiscovererStreamInfo *stream = l->data;
-		const GstTagList *stream_tags;
 
 		if (G_TYPE_CHECK_INSTANCE_TYPE (stream, GST_TYPE_DISCOVERER_AUDIO_INFO)) {
 			GstDiscovererAudioInfo *audio = (GstDiscovererAudioInfo*)stream;
@@ -1270,17 +1298,36 @@ discoverer_init_and_run (MetadataExtractor *extractor,
 		} else {
 			/* Unknown type - do nothing */
 		}
+	}
+
+	for (l = extractor->streams; l; l = g_list_next (l)) {
+		GstDiscovererStreamInfo *stream = l->data;
+		const GstTagList *stream_tags;
+		GstTagList *copy = NULL;
 
 		stream_tags = gst_discoverer_stream_info_get_tags (stream);
 
-		if (stream_tags) {
+		if (stream_tags)
+			copy = gst_tag_list_copy (stream_tags);
+
+		if (extractor->has_video &&
+		    copy &&
+		    gst_tag_list_get_tag_size (extractor->tagcache, "title") > 0)
+			gst_tag_list_remove_tag (copy, "title");
+
+		if (copy) {
 			gst_tag_list_insert (extractor->tagcache,
-					     stream_tags,
+					     copy,
 					     GST_TAG_MERGE_APPEND);
 		}
+
+		g_clear_pointer (&copy, gst_tag_list_unref);
 	}
 
 	gst_discoverer_info_unref (info);
+#if defined(HAVE_GSTREAMER_1_20)
+	gst_discoverer_stream_info_unref (sinfo);
+#endif
 
 	return TRUE;
 }
@@ -1319,7 +1366,10 @@ tracker_extract_gstreamer (const gchar          *uri,
 		}
 
 		if (extractor->toc == NULL) {
-			extractor->toc = tracker_cue_sheet_parse_uri (uri);
+			if (!local_conn)
+				local_conn = tracker_main_get_readonly_connection (NULL);
+
+			extractor->toc = tracker_cue_sheet_guess_from_uri (local_conn, uri);
 		}
 
 		if (extractor->toc == NULL &&
@@ -1410,9 +1460,11 @@ tracker_extract_module_init (GError **error)
 	/* Lifted from totem-video-thumbnailer */
 	const gchar *blocklisted[] = {
 		"bcmdec",
-		"fluiddec",
+		"fluidsynthmidi",
 		"vaapi",
-		"video4linux2"
+		"video4linux2",
+		"nvcodec",
+		"ges",
 	};
 	GstRegistry *registry;
 	guint i;
@@ -1423,10 +1475,20 @@ tracker_extract_module_init (GError **error)
 	for (i = 0; i < G_N_ELEMENTS (blocklisted); i++) {
 		GstPlugin *plugin =
 			gst_registry_find_plugin (registry,
-						  blocklisted[i]);
-		if (plugin)
+			                          blocklisted[i]);
+		if (plugin) {
+			g_debug ("Removing GStreamer plugin '%s' from registry",
+			         blocklisted[i]);
 			gst_registry_remove_plugin (registry, plugin);
+		}
 	}
 
 	return TRUE;
 }
+
+G_MODULE_EXPORT gboolean
+tracker_extract_module_shutdown (void)
+{
+	g_clear_object (&local_conn);
+	return TRUE;
+}
